################################################################
# Modifications beyond tkWWW 0.9 are Copyright 1993, 1994 by
# The University of Chicago Biological Sciences Division, Academic Computing
# Philip Stylianos
# Lee Newberg
################################################################

# selection.tcl

# This file controls the selection and the focus.  The selection
# is a range of the viewed page that is highlighted and is ready
# for clipboard operations (cut, copy) or deletion (Delete key).
# The focus is that part of the page where the insertion cursor
# is.
# 
# Note that the routines in this file handle the highlighting
# and location of the selection.  Most routines that look at the
# contents (i.e., cut, copy) of the selection are in
# clipboard.tcl.  However, the function phxW3SelectionDelete is
# in this file.

# Global variables
#   phxW3Selection: The widget which is or wholly contains the selection.
#   phxW3SelectionLeftOfFocus:
#   phxW3SelectionRightOfFocus: The focus is between these two
# widgets.  If the widgets are the same then the focus is
# strictly within the widget.  "Focus" is used interchangeably
# with "insertion mark."
#   phxW3SelectionLeftOfAnchor:
#   phxW3SelectionRightOfAnchor: The anchor is between these two
# widgets.  If the widgets are the same then the anchor is
# strictly within the widget.  The anchor is the location of the
# mouse-down event which defines one end of a selection being
# created by dragging the mouse.

# phxW3SelectionWidgetInitialize
# 
# Sets bindings for a selectable widget so that the selection
# routines will be invoked.

proc phxW3SelectionWidgetInitialize {w} {

    tkW3EditBind $w

    switch [phxW3WidgetClass $w] {
	para {
	    bind $w <Button-1> "phxW3SelectionParaButtonPress %W @%x,%y"

	    # Shift-Button-1: Extend selection from the anchor to
	    # the mouse-down point.
	    bind Text <Shift-Button-1> "tk_textSelectTo %W @%x,%y"
	}
	text -
	password -
	textarea {
	    bind $w <Button-1> "phxW3SelectionTextareaButtonPress %W @%x,%y"
	    bind Text <Shift-Button-1> "tk_textSelectTo %W @%x,%y"
	}
	entry {
	    bind $w <Button-1> "phxW3SelectionEntryButtonPress %W @%x"
	}
    }

    phxW3SelectionSetBinding $w <Shift-Right> {
	phxW3EditShiftArrowRight %W
	phxW3EditPickPlace
    }
    phxW3SelectionSetBinding $w <Shift-Left> {
	phxW3EditShiftArrowLeft %W
	phxW3EditPickPlace
    }
    phxW3SelectionSetBinding $w <Shift-Down> {
	phxW3EditShiftArrowDown %W
	phxW3EditPickPlace
    }
    phxW3SelectionSetBinding $w <Shift-Up> {
	phxW3EditShiftArrowUp %W
	phxW3EditPickPlace
    }

    # Promote <Double-1> class binding to a window binding.
    # This way *class's* <Double-1> will be invoked even though
    # *window's* <Button-1> is defined.

    phxW3SelectionSetBinding $w <Double-1> ""

    # Prepend to some of the window bindings
    phxW3SelectionSetBinding $w <Button-1> \
	"phxW3SelectionButtonPress %W %x %y"
    phxW3SelectionSetBinding $w <B1-Motion> \
	"phxW3SelectionMouseMotion %W %x %y"
    phxW3SelectionSetBinding $w <Shift-Button-1> \
	"phxW3SelectionMouseMotion %W %x %y"
    phxW3SelectionSetBinding $w <Shift-B1-Motion> \
	"phxW3SelectionMouseMotion %W %x %y"
}

# phxW3SelectionSetBinding {w e b}
# Puts binding $b ahead of the current binding for event $e of widget $w.

proc phxW3SelectionSetBinding {w e b} {
    set binding [bind $w $e]
    if {$binding == ""} { set binding [bind [winfo class $w] $e] }
    if {$binding == ""} { set binding [bind all $e] }
    if {$b != "" && $binding != ""} { append b "\n" }
    append b $binding
    bind $w $e $b
}

# phxW3SelectionClear
# 
# Low-level routine which de-selects everything.

proc phxW3SelectionClear {} {
    phxW3SelectionSet ""
    set phxW3SelectionLeftOfAnchor "none"
    set phxW3SelectionRightOfAnchor "none"

    # Also notify the windowing system.
    selection clear .
}

# phxW3SelectionDelete
# 
# Deletes everything selected.

proc phxW3SelectionDelete {} {

    # Remember the name of the widget to be deleted
    set selected [phxW3SelectionGet]
    if {$selected != ""} {
	set left [phxW3WidgetLeft $selected]
	set right [phxW3WidgetRight $selected]
	set class [phxW3WidgetClass $selected]

	switch $class {
	    entry {
		$selected delete sel.first sel.last
	    }
	    text -
	    password -
	    textarea {
		phxW3EditDeleteTextSelection $selected
	    }
	    default {
		if {[phxW3ModeGet] == "edit"} {
		    switch $class {
			para {
			    phxW3EditDeleteTextSelection $selected
			    if {[$selected compare 1.0 == end]} {
				phxW3SelectionSetFocus \
				    [phxW3WidgetLeft $selected] \
				    [set right [phxW3WidgetRight $selected]]
				destroy $selected
				phxW3WidgetWrap $right
			    } elseif {[$selected compare insert == 1.0]} {
				phxW3SelectionSetFocus \
				    [phxW3WidgetLeft $selected] $selected
			    } elseif {[$selected compare insert == end]} {
				phxW3SelectionSetFocus \
				    $selected [phxW3WidgetRight $selected]
			    }
			}
			image {
			    set parent [winfo parent $selected]
			    destroy $selected
			    if {[winfo children $parent] == ""} {
				destroy $parent
			    }
			    phxW3SelectionSetFocus $left $right
			    # Try to re-wrap.  Must use $right as an argument
			    # so that SelectionDelete of the first image works.
			    phxW3WidgetWrap $right
			}
			default {
			    # Handle Forms widgets and horizontal rules.
			    destroy $selected
			    phxW3SelectionSetFocus $left $right
			    # Try to re-wrap.  Must use $right as an argument
			    # so that SelectionDelete of the first image works.
			    phxW3WidgetWrap $right
			}
		    }
		}
	    }
	}
	phxW3SelectionClear
	if [string match ".f.*" $selected] {
		phxW3SaveDirtyBit
	}
    }    
}

# phxW3SelectionGet
# 
# Returns the name of the widget which owns the selection.

proc phxW3SelectionGet {} {
   global phxW3Selection
   return $phxW3Selection
}

# phxW3SelectionSet
# 
# Sets the name of the widget which owns the selection.

proc phxW3SelectionSet {w} {
   global phxW3Selection
   set phxW3Selection $w
}

# phxW3SelectionAll
# 
# Adds the entire document to the selection
# Not yet implemented!!

proc phxW3SelectionAll {} {
    tkW3OutputError "Select All: Not yet implemented"
}

# phxW3SelectionNone
#
# User routine which removes everything from the selection.

proc phxW3SelectionNone {} {
    phxW3SelectionClear
}

# phxW3SelectionButtonPress
#
# Sets a focus and starts a selection at the location of a mouse
# down event.

proc phxW3SelectionButtonPress {w x y} {

    # If something is currently selected then first de-select it.
    phxW3SelectionClear

    # If we fall within a text widget, let it handle everything
    set class [winfo class $w]
    if {$class != "Text" && $class != "Entry"} {
	# Figure out where to focus
	set width [winfo width $w]
	if {2 * $x < $width} {
	    phxW3SelectionSetFocus [phxW3WidgetLeft $w] $w
	} else {
	    phxW3SelectionSetFocus $w [phxW3WidgetRight $w]
	}
	# The focus is also the anchor for any selection formed
	# by dragging
	phxW3SelectionSetLeftOfAnchor [phxW3SelectionGetLeftOfFocus]
	phxW3SelectionSetRightOfAnchor [phxW3SelectionGetRightOfFocus]
    }

}

# phxW3SelectionParaButtonPress
# phxW3SelectionEntryButtonPress
# phxW3SelectionTextareaButtonPress
# 
# These control what happens when a button is pressed in a
# text-like widget.  There are, in some sense, two levels of
# focus.  In general, characters in paragraph text, inline
# images, rules, and entire elements of a form, are entities
# _between_ which the main focus can be set.  Alternatively, the
# focus can be placed _within_ a form text or textarea element.
# In the first case, the press of an arrow key will skip over an
# entire textarea element.  In the latter case, the press of an
# arrow key cannot move the cursor outside of the textarea
# element.

proc phxW3SelectionParaButtonPress {w index} {
    global tk_priv
    set tk_priv(selectMode) char
    $w mark set insert $index
    $w mark set anchor insert

    set left $w
    set right $w
    if {[$w compare insert == 1.0]} {set left [phxW3WidgetLeft $w]}
    if {[$w compare insert == end]} {set right [phxW3WidgetRight $w]}
    phxW3SelectionSetFocus $left $right
    phxW3SelectionSetLeftOfAnchor [phxW3SelectionGetLeftOfFocus]
    phxW3SelectionSetRightOfAnchor [phxW3SelectionGetRightOfFocus]
    if {[winfo class $w] == "Frame"} {
	phxW3CurrentFrame $w
    } else {
	phxW3CurrentFrame [winfo parent $w]
    }
}

proc phxW3SelectionEntryButtonPress {w index} {
    $w icursor $index
    $w select from $index

    phxW3SelectionSetFocus $w $w
    phxW3SelectionSetLeftOfAnchor $w
    phxW3SelectionSetRightOfAnchor $w
    if {[winfo class $w] == "Frame"} {
	phxW3CurrentFrame $w
    } else {
	phxW3CurrentFrame [winfo parent $w]
    }
}

proc phxW3SelectionTextareaButtonPress {w index} {
    global tk_priv
    set tk_priv(selectMode) char
    $w mark set insert $index
    $w mark set anchor insert

    phxW3SelectionSetFocus $w $w
    phxW3SelectionSetLeftOfAnchor $w
    phxW3SelectionSetRightOfAnchor $w
    if {[winfo class $w] == "Frame"} {
	phxW3CurrentFrame $w
    } else {
	phxW3CurrentFrame [winfo parent $w]
    }
}

# phxW3SelectionMouseMotion
# 
# Extends a selection according to the mouse motion.  For
# non-text widgets, mimics tk_textSelectTo to some degree.

proc phxW3SelectionMouseMotion {w x y} {

    if {![phxW3SelectionIsAnchored]} {
	# The anchor doesn't yet exist.  Better set it now.
	phxW3SelectionButtonPress $w $x $y
    }

    # Text widgets handle the selection by themselves.

    set class [winfo class $w]
    if {$class != "Text" && $class != "Entry"} {
	# Determine the current mouse location.
	set width [winfo width $w]
	if {2 * $x < $width} {
	    set indexLeft [phxW3WidgetLeft $w]
	    set indexRight $w
	} else {
	    set indexLeft $w
	    set indexRight [phxW3WidgetRight $w]
	}

	# Determine whether there is a single widget between the
	# anchor and the index.  Set the focus at the opposite
	# end of the selection from the anchor

	set anchorLeft [phxW3SelectionGetLeftOfAnchor]
	set anchorRight [phxW3SelectionGetRightOfAnchor]

	if {$indexRight == $anchorLeft} {
	    set new $indexRight
	    phxW3SelectionSetFocus $indexLeft $indexRight
	} elseif {$indexLeft == $anchorRight} {
	    set new $indexLeft
	    phxW3SelectionSetFocus $indexLeft $indexRight
	} else {
	    set new ""
	    phxW3SelectionSetFocus $anchorLeft $anchorRight
	}
	
	# Look for a change
	set old [phxW3SelectionGet]
	if {$new == "" && $old != ""} {
	    phxW3SelectionClear
	} elseif {$old == "" && $new != ""} {
	    phxW3SelectionSet $new
	    global phxW3SelectionColor
	    $new config -bg $phxW3SelectionColor
	    catch {$new config -activebackground $phxW3SelectionColor}
	    selection own $new "phxW3ModeConfigure $new"
	}
    } else {
	phxW3SelectionSet $w
    }
}

# phxW3SelectionClearFocus
proc phxW3SelectionClearFocus {} {
    global phxW3SelectionLeftOfFocus
    global phxW3SelectionRightOfFocus
    set phxW3SelectionLeftOfFocus "none"
    set phxW3SelectionRightOfFocus "none"
    # Also notify the windowing system
    focus none
}

# phxW3SelectionIsFocused
#
# Returns TRUE iff the focus is set.

proc phxW3SelectionIsFocused {} {
    global phxW3SelectionLeftOfFocus
    global phxW3SelectionRightOfFocus
    return [expr {$phxW3SelectionLeftOfFocus != "none" && \
	$phxW3SelectionRightOfFocus != "none"}]
}

# phxW3SelectionGetLeftOfFocus
# 
# Returns the widget to the left of the insertion point.  If the
# insertion point is within a para widget at any position other
# than 1.0, then the text widget will be returned.

proc phxW3SelectionGetLeftOfFocus {} {
    global phxW3SelectionLeftOfFocus
    return $phxW3SelectionLeftOfFocus
}

# phxW3SelectionGetRightOfFocus
# 
# Returns the widget to the right of the insertion point.  If
# the insertion point is within a para widget at any position
# other than end, then the text widget will be returned.

proc phxW3SelectionGetRightOfFocus {} {
    global phxW3SelectionRightOfFocus
    return $phxW3SelectionRightOfFocus
}

# Setting the focus
# 
# Important!  If you're changing both the LeftOfFocus _and_ the
# RightOfFocus, then you should _not_ call
	# phxW3SelectionSetFocus unless you are giving it both new
# values.

# phxW3SelectionSetFocus

proc phxW3SelectionSetFocus {left {right ""}} {
    global phxW3SelectionLeftOfFocus
    global phxW3SelectionRightOfFocus

    if {$right == ""} {
	set right $left
    }

    if {$phxW3SelectionLeftOfFocus != $left || \
	$phxW3SelectionRightOfFocus != $right} \
    {
	if {$left == "none" && $right == "none"} {
	    phxW3SelectionClearFocus
	} else {
	    set phxW3SelectionLeftOfFocus $left
	    set phxW3SelectionRightOfFocus $right
	    phxW3SelectionSetFocusInsert
	}
    }
}

# phxW3SelectionSetFocusInsert
# 
# Makes sure the insertion cursor is displayed correctly.  If
# we're in or adjacent to para text, put the focus in the para
# widget.  If we're within a text, password or textarea widget
# then put the focus there.  Otherwise, put the focus with the
# widget to the right -- be sure that this widget does show the
# insertion cursor.

proc phxW3SelectionSetFocusInsert {} {
    global phxW3SelectionLeftOfFocus
    global phxW3SelectionRightOfFocus

    set classLeft [phxW3WidgetClass $phxW3SelectionLeftOfFocus]
    set classRight [phxW3WidgetClass $phxW3SelectionRightOfFocus]

    if {$phxW3SelectionLeftOfFocus == $phxW3SelectionRightOfFocus} {
	set focus $phxW3SelectionLeftOfFocus
	set class $classLeft
    } elseif {$classLeft == "para"} {
	set focus $phxW3SelectionLeftOfFocus
	$phxW3SelectionLeftOfFocus mark set insert end
	set class $classLeft
    } elseif {$classRight != "end"} {
	set focus $phxW3SelectionRightOfFocus
	if {$classRight == "para"} {
	    $phxW3SelectionRightOfFocus mark set insert 1.0
	}
	set class $classRight
    } elseif {$classLeft != "start"} {
	set focus $phxW3SelectionLeftOfFocus
	set class $classLeft
    } else {
	set focus .f
	set class none
    }

    focus $focus

    # Hide the insertion cursor if it is between a form text, password, 
    # or textarea widget rather than in it.
    switch $class {
	text -
	password -
	textarea {
	    if {$phxW3SelectionLeftOfFocus != $phxW3SelectionRightOfFocus} {
		$phxW3SelectionRightOfFocus configure -insertwidth 0
	    } else {
		$phxW3SelectionRightOfFocus configure -insertwidth 2
	    }
	}
    }

    # Allow the Paste menu item only when it editing is allowed
    # where the focus is.
    # Editing is allowed if either of the following is true.
    # 1. We're in edit mode and the focus is a descendent of ".f".
    # 2. The focus is in an Entry or Text widget with "-state normal".
    #    (This happens when we're in a dialogue box.)

    set Xclass [winfo class $focus]
    if {([phxW3ModeGet] == "edit" && [regexp {^\.f} $focus]) || \
	(($Xclass == "Text" || $Xclass == "Entry") && \
	 [lindex [$focus configure -state] 4] == "normal")} {
	.menu.edit.m entryconfigure Paste -state normal
    } else {
	.menu.edit.m entryconfigure Paste -state disabled
    }
}


# phxW3SelectionIsAnchored
# 
# Returns TRUE iff the anchor is set.  The anchor point is the
# location of the mouse-down event at the beginning of a drag
# which defines a selection.

proc phxW3SelectionIsAnchored {} {
    global phxW3SelectionLeftOfAnchor
    global phxW3SelectionRightOfAnchor
    return [expr {$phxW3SelectionLeftOfAnchor != "none" && \
	$phxW3SelectionRightOfAnchor != "none"}]
}

# phxW3SelectionGetLeftOfAnchor
# 
# Returns the widget to the right of the anchor point.  If the
# anchor point is within a text widget at any position other
# than end, then the text widget will be returned.

proc phxW3SelectionGetLeftOfAnchor {} {
    global phxW3SelectionLeftOfAnchor
    return $phxW3SelectionLeftOfAnchor
}

# phxW3SelectionGetRightOfAnchor
# 
# Returns the widget to the left of the anchor point.  If the
# anchor point is within a text widget at any position other
# than 1.0, then the text widget will be returned.  The anchor
# point is the location of the mouse-down event at the beginning
# of a drag which defines a selection.

proc phxW3SelectionGetRightOfAnchor {} {
    global phxW3SelectionRightOfAnchor
    return $phxW3SelectionRightOfAnchor
}

# phxW3SelectionSetLeftOfAnchor

proc phxW3SelectionSetLeftOfAnchor {w} {
    global phxW3SelectionLeftOfAnchor
    set phxW3SelectionLeftOfAnchor $w
}

# phxW3SelectionSetRightOfAnchor

proc phxW3SelectionSetRightOfAnchor {w} {
    global phxW3SelectionRightOfAnchor
    set phxW3SelectionRightOfAnchor $w
}

# tk_textSelectTo

# A modification of the default tk_textSelectTo.  Our version sets the
# insert mark to the the mouse-up location rather than the
# mouse-down location.

proc tk_textSelectTo {w index} {
    global tk_priv

    if [catch {$w index anchor}] {
	$w mark set anchor $index
    }
    case $tk_priv(selectMode) {
	char {
	    if [$w compare $index < anchor] {
		set first $index
		set last anchor
		$w mark set insert $first
	    } else {
		set first anchor
		set last [$w index $index+1c]
		$w mark set insert $last
	    }
	}
	word {
	    if [$w compare $index < anchor] {
		set first [$w index "$index wordstart"]
		set last [$w index "anchor wordend"]
		$w mark set insert $first
	    } else {
		set first [$w index "anchor wordstart"]
		set last [$w index "$index wordend"]
		$w mark set insert $last
	    }
	}
	line {
	    if [$w compare $index < anchor] {
		set first [$w index "$index linestart"]
		set last [$w index "anchor lineend + 1c"]
		$w mark set insert $first
	    } else {
		set first [$w index "anchor linestart"]
		set last [$w index "$index lineend + 1c"]
		$w mark set insert $last
	    }
	}
    }
    $w tag remove sel 0.0 $first
    $w tag add sel $first $last
    $w tag remove sel $last end
    if {[$w compare $first != $last]} {
	phxW3SelectionSet $w
    } else {
	phxW3SelectionClear
    }
}

proc phxW3SelectionInitialize {} {
    global phxW3Selection
    global phxW3SelectionLeftOfFocus
    global phxW3SelectionRightOfFocus
    global phxW3SelectionLeftOfAnchor
    global phxW3SelectionRightOfAnchor

    set phxW3Selection ""
    set phxW3SelectionLeftOfFocus "none"
    set phxW3SelectionRightOfFocus "none"
    set phxW3SelectionLeftOfAnchor "none"
    set phxW3SelectionRightOfAnchor "none"
}
